/*
 * vf_mezzmo.c: 
 * Mezzmo Filter for FFmpeg by Christopher Milowka, Conceiva Pty.Ltd.
 * ---------------------------------------------------------------------
 * 
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */



/**
 * @file
 * Mezzmo Filter
 */
#include <stdio.h>
#include <stdlib.h>                                // malloc(), free()
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <dirent.h>
#include "config.h"
#include "avfilter.h"
#include "libavutil/opt.h"
#include "libavutil/parseutils.h"
#include "libavutil/colorspace.h"
#include "libavutil/pixdesc.h"
#include "libavutil/time.h"                // av_gettime(void)
#include "internal.h"
#include "video.h"

#define _LFS64_LARGEFILE        0
#define _LARGEFILE64_SOURCE        0
#include "zlib.h"                                // ZLIB


extern int64_t mezzmo_video_duration;            // Already calculated

#define SET_PIXEL(pic_ref, yuv_color, x, y, hsub, vsub) { \
    luma_pos    = (x        ) + (y        )*pic_ref->linesize[0]; \
    chroma_pos1 = (x >> hsub) + (y >> vsub)*pic_ref->linesize[1]; \
    chroma_pos2 = (x >> hsub) + (y >> vsub)*pic_ref->linesize[2]; \
    pic_ref->data[0][luma_pos]    = (yuv_color[3]*yuv_color[0] + (255 - yuv_color[3])*pic_ref->data[0][luma_pos])    >> 8; \
    pic_ref->data[1][chroma_pos1] = (yuv_color[3]*yuv_color[1] + (255 - yuv_color[3])*pic_ref->data[1][chroma_pos1]) >> 8; \
    pic_ref->data[2][chroma_pos2] = (yuv_color[3]*yuv_color[2] + (255 - yuv_color[3])*pic_ref->data[2][chroma_pos2]) >> 8; \
}

#define GET_PIXEL(pic_ref, yuv_color, x, y, hsub, vsub) { \
    yuv_color[0] = pic_ref->data[0][ (x        ) + (y        )*pic_ref->linesize[0] ]; \
    yuv_color[1] = pic_ref->data[1][ (x >> hsub) + (y >> vsub)*pic_ref->linesize[1] ]; \
    yuv_color[2] = pic_ref->data[2][ (x >> hsub) + (y >> vsub)*pic_ref->linesize[2] ]; \
}

#define rgb2y(R,G,B)  ( (( 263*R + 516*G + 100*B) >> 10) + 16  )
#define rgb2u(R,G,B)  ( ((-152*R - 298*G + 450*B) >> 10) + 128 )
#define rgb2v(R,G,B)  ( (( 450*R - 376*G -  73*B) >> 10) + 128 )

#define MAX_MEZZMOITEM    4

enum { Y, U, V, A };

typedef unsigned char yuva_t[4];
typedef unsigned char bool;

typedef struct {
    unsigned char   mode;                    
    int16_t         b_xl;                    
    int16_t         b_yu;                    
    int16_t         b_wd;                    
    int16_t         b_ht;                    
    int64_t         ptsS;                    
    int64_t         ptsE;                    
    uint8_t         fade;                    
    uint8_t         roll;                    
    int             jobF;                    
    int             jobR;                    
    uint32_t        uzip;    
    uint8_t         flag;
    unsigned char * area;                    
} MezzmoItem;



typedef struct {
    const AVClass * class;                            
    char          * fifoPath;                        
    int             ibox;    
    int             dSbs;
    char          * boxR_char;                        
    char          * boxG_char;    
    char          * bkgr_char;    
    FILE          * fifoFile;                        
    int             hsub, vsub;
    int16_t         hsbs, vsbs;                    
    unsigned char   boxR_yuva[4];                    
    unsigned char   boxG_yuva[4];                    
    unsigned char   bkgr_yuva[4];                    
    MezzmoItem      mezzmoItem[MAX_MEZZMOITEM];        
    uint8_t         fifoDone;
} MezzmoContext;


#define OFFSET(x) offsetof(MezzmoContext, x)
#define FLAGS AV_OPT_FLAG_FILTERING_PARAM|AV_OPT_FLAG_VIDEO_PARAM

static const AVOption mezzmo_options[]= {
{"i",  "set Mezzmo fifo", OFFSET(fifoPath)    , FF_OPT_TYPE_STRING, {.str=NULL}    ,  CHAR_MIN, CHAR_MAX, FLAGS },
{"r",  "set start color", OFFSET(boxR_char)   , FF_OPT_TYPE_STRING, {.str=NULL}    ,  CHAR_MIN, CHAR_MAX, FLAGS },
{"g",  "set end color"  , OFFSET(boxG_char)   , FF_OPT_TYPE_STRING, {.str=NULL}    ,  CHAR_MIN, CHAR_MAX, FLAGS },
{"p",  "set progress"   , OFFSET(ibox)        , FF_OPT_TYPE_INT   , {.dbl=0    }   ,         0,        1, FLAGS },
{"d",  "set dimensions" , OFFSET(dSbs)        , FF_OPT_TYPE_INT   , {.dbl=0    }   ,         0,        2, FLAGS },
{"u",  "set box"        , OFFSET(bkgr_char)   , FF_OPT_TYPE_STRING, {.str=NULL}    ,  CHAR_MIN, CHAR_MAX, FLAGS },
{NULL},
};

static const char *mezzmo_get_name(void *ctx)
{
    return "Mezzmo Filter";
}




static int query_formats(AVFilterContext *ctx)
{
    static const enum PixelFormat pix_fmts[] = {
        PIX_FMT_YUV444P,  
        PIX_FMT_YUV422P,  
        PIX_FMT_YUV420P,
        PIX_FMT_YUV411P,  
        PIX_FMT_YUV410P,
        PIX_FMT_YUVJ444P, 
        PIX_FMT_YUVJ422P, 
        PIX_FMT_YUVJ420P,
        PIX_FMT_YUV440P,  
        PIX_FMT_YUVJ440P,
        PIX_FMT_NONE
    };
    ff_set_common_formats(ctx, ff_make_format_list(pix_fmts));
    return 0;
}


static inline int extract_color(AVFilterContext *ctx, char *color_str, unsigned char *color)
{
    uint8_t err;
    if ((err = av_parse_color(color, color_str, -1, ctx))) { return err; }    
    return 0;
}



static av_cold int init( AVFilterContext *ctx)
{
    MezzmoContext * mezzmoCtx = ctx->priv;
    int      err = 0;
    int      pos = 0;
    uint8_t  rgba[4];    
        //
    mezzmoCtx->hsub  = 0;
    mezzmoCtx->vsub  = 0;
    mezzmoCtx->hsbs  = 0;
    mezzmoCtx->vsbs  = 0;
    //mezzmoCtx->ibox  = 1;
    //mezzmoCtx->dSbs  = 0;
    //mezzmoCtx->fifoPath = 0;
    mezzmoCtx->fifoFile = 0;
    mezzmoCtx->fifoDone = 0;
	
	if (mezzmoCtx->boxR_char == NULL)
	{
		mezzmoCtx->boxR_char = av_strdup("");
	}
	
    if (mezzmoCtx->boxG_char == NULL)
	{
		mezzmoCtx->boxG_char = av_strdup("");
	}
	
    if (mezzmoCtx->bkgr_char == NULL)
	{
		mezzmoCtx->bkgr_char = av_strdup("");
	}
	
    for (pos=0; pos < MAX_MEZZMOITEM; pos++)
    {
        mezzmoCtx->mezzmoItem[pos].mode = 0;
        mezzmoCtx->mezzmoItem[pos].b_xl = 0;
        mezzmoCtx->mezzmoItem[pos].b_yu = 0;
        mezzmoCtx->mezzmoItem[pos].b_wd = 0;
        mezzmoCtx->mezzmoItem[pos].b_ht = 0;
        mezzmoCtx->mezzmoItem[pos].ptsS = 0;
        mezzmoCtx->mezzmoItem[pos].ptsE = 0;
        mezzmoCtx->mezzmoItem[pos].fade = 0;
        mezzmoCtx->mezzmoItem[pos].roll = 0;
        mezzmoCtx->mezzmoItem[pos].area = 0;
        mezzmoCtx->mezzmoItem[pos].uzip = 0;
        mezzmoCtx->mezzmoItem[pos].flag = 0;
    }
   /* av_opt_set_defaults( mezzmoCtx );
    if ((err = (av_set_options_string(mezzmoCtx, args, "=", "|"))) < 0) 
    {
        av_log(ctx, AV_LOG_ERROR, "Error: Parsing options string: '%s'\n", args);
    }*/
        // Report Mezzmo Fifo File
    if (mezzmoCtx->fifoPath) 
    {
        mezzmoCtx->fifoFile = av_fopen_utf8(mezzmoCtx->fifoPath, "rb");
        if (!(mezzmoCtx->fifoFile)) 
        {
            av_log(NULL, AV_LOG_ERROR, "Error: Mezzmo fifo file '%s' could not be opened.\n", mezzmoCtx->fifoPath);
        } 
        else
        {
            av_log(NULL, AV_LOG_INFO, "Input #0, Mezzmo Fifo, from '%s'.\n", mezzmoCtx->fifoPath);
            mezzmoCtx->fifoDone = 1;
        }
    }//else Mezzmo fifo not provided
        // Report Mezzmo Progress Bar
    if (mezzmoCtx->ibox) 
    {
        av_log(NULL, AV_LOG_INFO, "Input #0, Mezzmo Progress Bar S=%s E=%s.\n", mezzmoCtx->boxR_char, mezzmoCtx->boxG_char);
    }
        // Convert colors
    if (mezzmoCtx->boxR_char && (strlen(mezzmoCtx->boxR_char) >= 8))
    {
        err = extract_color( ctx, mezzmoCtx->boxR_char, rgba );
    }

    if (!mezzmoCtx->boxR_char || !(mezzmoCtx->boxR_char[0]) || err ) 
    {
        rgba[0] = 0x8A;    rgba[1] = 0x44; rgba[2] = 0x44;    rgba[3] = 0x88;
    }
    else if (strlen(mezzmoCtx->boxR_char) < 4)     rgba[3] = 0xcc;

    mezzmoCtx->boxR_yuva[0] = rgb2y(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->boxR_yuva[1] = rgb2u(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->boxR_yuva[2] = rgb2v(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->boxR_yuva[3] = (255 - rgba[3]);

    if (mezzmoCtx->boxG_char && (strlen(mezzmoCtx->boxG_char) >= 8))
    {
        err = extract_color( ctx, mezzmoCtx->boxG_char, rgba );
    }

    if (!mezzmoCtx->boxG_char || !(mezzmoCtx->boxG_char[0]) || err )
    {
        rgba[0] = 0x57;    rgba[1] = 0x8B; rgba[2] = 0x57;    rgba[3] = 0x88;
    }

    mezzmoCtx->boxG_yuva[0] = rgb2y(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->boxG_yuva[1] = rgb2u(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->boxG_yuva[2] = rgb2v(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->boxG_yuva[3] = (255 - rgba[3]);
    if (mezzmoCtx->bkgr_char && (strlen(mezzmoCtx->bkgr_char) >= 8))
    {
        err = extract_color( ctx, mezzmoCtx->bkgr_char, rgba );
    }
    if (!mezzmoCtx->bkgr_char || !(mezzmoCtx->bkgr_char[0]) || err ) 
    {
        rgba[0] = 0x00;    rgba[1] = 0x00; rgba[2] = 0x00;    rgba[3] = 0xFF;
    }
    else if (strlen(mezzmoCtx->bkgr_char) < 4)     rgba[3] = 0xFF;
    mezzmoCtx->bkgr_yuva[0] = rgb2y(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->bkgr_yuva[1] = rgb2u(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->bkgr_yuva[2] = rgb2v(rgba[0], rgba[1], rgba[2]);
    mezzmoCtx->bkgr_yuva[3] = (0xFF - rgba[3]);

    return 0;
}



static av_cold void uninit(AVFilterContext *ctx)
{
    MezzmoContext * mezzmoCtx = ctx->priv;    
    int pos = 0;
    if (mezzmoCtx->fifoFile)  fclose ( mezzmoCtx->fifoFile );
    if (mezzmoCtx->fifoPath)  av_free( mezzmoCtx->fifoPath );
    if (mezzmoCtx->boxR_char) av_free( mezzmoCtx->boxR_char);
    if (mezzmoCtx->boxG_char) av_free( mezzmoCtx->boxG_char);
    for (pos=0; pos < MAX_MEZZMOITEM; pos++)
    {
        if (mezzmoCtx->mezzmoItem[pos].area) av_free( mezzmoCtx->mezzmoItem[pos].area );
    }
}



static int config_input(AVFilterLink *inlink)
{
    MezzmoContext *mezzmoCtx = inlink->dst->priv;
    mezzmoCtx->hsub = av_pix_fmt_descriptors[inlink->format].log2_chroma_w;
    mezzmoCtx->vsub = av_pix_fmt_descriptors[inlink->format].log2_chroma_h;
    return 0;
}



static bool draw_box(AVFrame *pic_ref, int v_wd, int v_ht 
                     , int16_t v_xl, int16_t v_yu, int16_t b_wd, int16_t b_ht 
                     , unsigned char yuva[4], bool isFill, int hsub, int vsub, int d_xl, int d_yu )
{
    const int maxBrd = 2;
    int16_t i, j;
        int16_t min_i,min_j,max_i,max_j;
    int xo = maxBrd-1;    int yo = maxBrd;
    min_i = (v_xl < 0) ? 0 : v_xl;
    min_j = (v_yu < 0) ? 0 : v_yu;
    max_i = v_xl + b_wd;    if (max_i > v_wd) max_i = v_wd;
    max_j = v_yu + b_ht;    if (max_j > v_ht) max_j = v_ht;
    min_i += d_xl;    max_i += d_xl;
    min_j += d_yu;    max_j += d_yu;    
        // Draw Box by YUV
    if (yuva[3] != 0xFF) 
    {    // Slower draw with alpha transparency
        unsigned int luma_pos, chroma_pos1, chroma_pos2;
        for (j = min_j; j < max_j; j++) 
        {
            for (i = min_i; i < max_i; i++)
            {
                SET_PIXEL(pic_ref, yuva, i, j, hsub, vsub);
            }
        }
    } 
    else if (isFill)
    {    // Faster draw box without alpha trasnparency
        int    plane;
        uint8_t * pos;
        for (plane = 0; plane < 3 && pic_ref->data[plane]; plane++) 
        {
            int hsub1 = plane == 1 || plane == 2 ? hsub : 0;
            int vsub1 = plane == 1 || plane == 2 ? vsub : 0;
            pos = pic_ref->data[plane] + (min_j >> vsub1) * pic_ref->linesize[plane] + (min_i >> hsub1);
            for (i = 0; i < (b_ht >> vsub1); i++) 
            {
                memset( pos, yuva[plane], (b_wd >> hsub1) );
                pos += pic_ref->linesize[plane];
            }
        }
    }
    else
    {    // Faster draw rectangle without alpha transparency
        int plane;                
        unsigned char ** row = av_malloc(maxBrd*sizeof(unsigned char*));
        for (j = min_j; j < max_j; j++) 
        {
            row[0] = pic_ref->data[0] + j * pic_ref->linesize[0];
            for (plane = 1; plane < xo; plane++)
            {
                row[plane] = pic_ref->data[plane] + pic_ref->linesize[plane] * (j >> vsub);
            }
            for (i = min_i; i < max_i; i++)
            {
                if ((j - min_j < xo) || (max_j - j < yo) 
                 || (i - min_i < xo) || (max_i - i < yo))
                {
                    row[0][i] = yuva[0];
                    row[1][i >> hsub] = yuva[1];
                    row[2][i >> hsub] = yuva[2];
                }
            }
        }
        av_free(row);
        }
    return 1;
}



static int process_ibox( AVFilterContext * ctx, AVFrame * pic_ref, int64_t vpts, double timeBase )
{
    static int64_t time_set = -1;
    static int64_t time_job = -1;
    static int16_t i_wd = -1;
    MezzmoContext * mezzmoCtx = ctx->priv;
    int v_wd = (0 != mezzmoCtx->hsbs) ? mezzmoCtx->hsbs : pic_ref->width;
    int v_ht = (0 != mezzmoCtx->vsbs) ? mezzmoCtx->vsbs : pic_ref->height;
    int16_t p_wd = (v_wd >> 3);
    int16_t p_ht =  2;
    int16_t p_xl = (v_wd - p_wd - 20);
    int16_t p_yu = 10;
    int16_t a_wd;
    int64_t time_cur = 0;
    float perc = 0.0;
    float fact = 0.0;
    if (!mezzmoCtx->ibox) return 0;
    if (p_xl < 0) return 0;
	
    if (time_job < 0) time_job = (mezzmo_video_duration * 1000) / AV_TIME_BASE;
    if (time_job <= 0) return 0;
    if (time_set < 0) time_set = av_gettime();    
    time_cur = av_gettime() - time_set;
    time_cur = time_cur / 1000;    
    if (time_cur > 0)
    {
        fact = vpts / time_cur;
        perc = ((fact * vpts) / time_job);
    }
    if (perc > 1.0)
    {
        mezzmoCtx->ibox = 0;
        return 0;
    }    
    draw_box( pic_ref, v_wd, v_ht, p_xl, p_yu, p_wd, p_ht, mezzmoCtx->boxR_yuva, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, 0 );
    if (0 != mezzmoCtx->hsbs) draw_box( pic_ref, v_wd, v_ht, p_xl, p_yu, p_wd, p_ht, mezzmoCtx->boxR_yuva, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, mezzmoCtx->hsbs, 0 );
    if (0 != mezzmoCtx->vsbs) draw_box( pic_ref, v_wd, v_ht, p_xl, p_yu, p_wd, p_ht, mezzmoCtx->boxR_yuva, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, mezzmoCtx->vsbs );
    a_wd = (p_wd * perc) + 0.5;
    if (a_wd > i_wd) i_wd = a_wd;
    if (i_wd > 0) 
    {
        draw_box( pic_ref, v_wd, v_ht, p_xl, p_yu, i_wd, p_ht, mezzmoCtx->boxG_yuva, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, 0 );
        if (0 != mezzmoCtx->hsbs) draw_box( pic_ref, v_wd, v_ht, p_xl, p_yu, i_wd, p_ht, mezzmoCtx->boxG_yuva, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, mezzmoCtx->hsbs, 0 );
        if (0 != mezzmoCtx->vsbs) draw_box( pic_ref, v_wd, v_ht, p_xl, p_yu, i_wd, p_ht, mezzmoCtx->boxG_yuva, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, mezzmoCtx->vsbs );
    }
    return 1;
}



static inline bool draw_img( AVFrame *pic_ref, int v_wd, int v_ht
                            , int16_t b_xl, int16_t b_yu, int16_t b_wd, int16_t b_ht 
                            , unsigned char * area, uint8_t a_as, int hsub, int vsub, int d_xl, int d_yu )
{
    int16_t s_xl,s_yu,i_xl,i_yu,o_xl,min_i,min_j,max_i,max_j,a_at;
    uint8_t y,u,v,a,A; 
    unsigned int luma_pos, chroma_pos1, chroma_pos2;
    if (!area)    return 1;
    if (!pic_ref) return 1;
        //
    min_i = (b_xl < 0) ? 0 : b_xl;
    min_j = (b_yu < 0) ? 0 : b_yu;
    max_i = b_xl + b_wd;    if (max_i > v_wd) max_i = v_wd;
    max_j = b_yu + b_ht;    if (max_j > v_ht) max_j = v_ht;
    i_xl = o_xl = min_i - b_xl;
    i_yu = min_j - b_yu;
    min_i += d_xl;    max_i += d_xl;
    min_j += d_yu;    max_j += d_yu;    
    for (s_yu = min_j; s_yu < max_j; s_yu++, i_yu++) 
    {
        i_xl = o_xl;
        for (s_xl = min_i; s_xl < max_i; s_xl++, i_xl++)
        {
            unsigned int p = i_yu * (b_wd << 2) + (i_xl << 2);
            y = *(area + p + 0);
            u = *(area + p + 1);
            v = *(area + p + 2);
            a = *(area + p + 3);
            a_at = a - a_as;
            a = (a_at < 0) ? 0 : a_at;
            luma_pos    = (s_xl        ) + (s_yu        ) * pic_ref->linesize[0]; 
            chroma_pos1 = (s_xl >> hsub) + (s_yu >> vsub) * pic_ref->linesize[1]; 
            chroma_pos2 = (s_xl >> hsub) + (s_yu >> vsub) * pic_ref->linesize[2];                 
            A = 255 - a;
            if (0xff == a)
            {    // Opaque pixel
                pic_ref->data[0][luma_pos] = y;
                pic_ref->data[1][chroma_pos1] = u;
                pic_ref->data[2][chroma_pos2] = v;
            }
            else if (0x00 != a)
            {    // Blended
                pic_ref->data[0][luma_pos]    = (a*y + A*pic_ref->data[0][luma_pos])    >> 8; 
                pic_ref->data[1][chroma_pos1] = (a*u + A*pic_ref->data[1][chroma_pos1]) >> 8; 
                pic_ref->data[2][chroma_pos2] = (a*v + A*pic_ref->data[2][chroma_pos2]) >> 8; 
            }//else Completely transparent pixel, nothing to do
        }
    }
    return 1;
}


static bool load_item(MezzmoContext *mezzmoCtx, AVFrame *pic_ref,MezzmoItem * mezzmoItem)
{
    uint8_t roll = 0;
    long    maxData = 0;
    int     posItem = 0;
    uLongf  maxUzip = 0;
    int     retUZip = 0;
    unsigned char * outUzip = 0;
    int v_wd = pic_ref->width;
    int v_ht = pic_ref->height;
    if (0 != mezzmoCtx->hsbs) v_wd = mezzmoCtx->hsbs;
    if (0 != mezzmoCtx->vsbs) v_ht = mezzmoCtx->vsbs;
        //
    if (!mezzmoItem) return 0;    
    if (mezzmoItem->area) return  0;
    if (!mezzmoCtx->fifoFile) return 0;
    memset( mezzmoItem, 0, sizeof(MezzmoItem) ); 
    if (fread( &(mezzmoItem->mode), 1, 1, mezzmoCtx->fifoFile) != 1) return 0;
    posItem = mezzmoItem->mode - 1;
    if (posItem < 0) return 0;
    if (posItem >= MAX_MEZZMOITEM) return 0;
    mezzmoItem->mode = posItem;
    if (fread( &(mezzmoItem->b_xl), 2, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (fread( &(mezzmoItem->b_yu), 2, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (fread( &(mezzmoItem->b_wd), 2, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (fread( &(mezzmoItem->b_ht), 2, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (fread( &(mezzmoItem->ptsS), 8, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (fread( &(mezzmoItem->ptsE), 8, 1, mezzmoCtx->fifoFile) != 1) return 0;    
    if (fread( &(mezzmoItem->fade), 1, 1, mezzmoCtx->fifoFile) != 1) return 0;    
    if (fread( &(mezzmoItem->roll), 1, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (fread( &(mezzmoItem->uzip), 4, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (fread( &(mezzmoItem->flag), 1, 1, mezzmoCtx->fifoFile) != 1) return 0;
    if (mezzmoItem->b_wd < 8) return 0;
    if (mezzmoItem->b_ht < 8) return 0;
    if (mezzmoItem->ptsS < 0) return 0;
    if (mezzmoItem->ptsE < 0) return 0;
    if (mezzmoItem->ptsE < mezzmoItem->ptsS) return 0;
    mezzmoItem->ptsS *= 10;
    mezzmoItem->ptsE *= 10;
    maxData = mezzmoItem->b_wd * mezzmoItem->b_ht * sizeof(yuva_t);
    if (maxData < 32) return 0; 
    mezzmoItem->area = (unsigned char *) av_malloc( maxData );
    if (NULL == mezzmoItem->area) return 0;
    if (mezzmoItem->uzip > 0)
    {
        retUZip = Z_MEM_ERROR;
        outUzip = (unsigned char *) av_malloc( mezzmoItem->uzip );
        if (outUzip != NULL)
        {
            retUZip = Z_STREAM_ERROR;
            if (fread( outUzip, 1, mezzmoItem->uzip, mezzmoCtx->fifoFile) == mezzmoItem->uzip)
            {
                maxUzip = maxData;
                retUZip = uncompress((Bytef*)mezzmoItem->area, &maxUzip, (const Bytef*)outUzip, mezzmoItem->uzip );
                if (maxUzip != maxData) retUZip = Z_DATA_ERROR;
            }
        }
        if (retUZip != Z_OK)
        {
            if (outUzip) { av_free( outUzip ); outUzip = 0; }
            if (mezzmoItem->area) { av_free( mezzmoItem->area );  mezzmoItem->area = 0; }
            return 0;
        }
        av_free( outUzip );
    }
    else
    {
        if (fread( mezzmoItem->area, 1, maxData, mezzmoCtx->fifoFile) != maxData) return 0;
    }
    mezzmoItem->jobF = 0;
    if (mezzmoItem->fade > 0) mezzmoItem->jobF = 255;
    mezzmoItem->jobR = 0;
    if (mezzmoItem->roll > 0)
    {
        roll = mezzmoItem->roll;
             if (1 == roll) mezzmoItem->jobR = -(mezzmoItem->b_wd);
        else if (2 == roll) mezzmoItem->jobR =  v_wd;
        else if (3 == roll) mezzmoItem->jobR = -(mezzmoItem->b_ht);
        else if (4 == roll) mezzmoItem->jobR =  v_ht;
    }
    return 1;
}


static bool load_fifo(MezzmoContext *mezzmoCtx, AVFrame *pic_ref)
{
    MezzmoItem * mezzmoItem;
    int posItem = 0;
    int v_wd = pic_ref->width;
    int v_ht = pic_ref->height;
    if (0 != mezzmoCtx->hsbs) v_wd = mezzmoCtx->hsbs;
    if (0 != mezzmoCtx->vsbs) v_ht = mezzmoCtx->vsbs;
        //
    if (!mezzmoCtx->fifoFile) return 0;
    if (!pic_ref) return 0;
    for (posItem = 0; posItem < MAX_MEZZMOITEM; posItem++ )
    {    
        mezzmoItem = &(mezzmoCtx->mezzmoItem[posItem]);
        if (0 == mezzmoItem->area)
        {
            if (!(load_item(mezzmoCtx, pic_ref, mezzmoItem)))
            {
                if (mezzmoItem->area)
                {
                    if (mezzmoItem->area) { av_free( mezzmoItem->area );  mezzmoItem->area = 0; }
                }
                memset( mezzmoItem, 0, sizeof(MezzmoItem) ); 
            }
            else if (mezzmoItem->flag >= 30)
            {
                if (0 == mezzmoItem->b_xl)
                {
                    mezzmoItem->b_xl = (v_wd - mezzmoItem->b_wd) >> 1;
                }
                mezzmoItem->b_yu = v_ht - mezzmoItem->b_yu;
                if (mezzmoItem->b_yu < 0) mezzmoItem->b_yu = 20;
            }
        }
    }
    return 1;
}



static bool process_fifo( AVFilterContext *ctx, AVFrame *pic_ref, int64_t vpts, double timeBase)
{
    bool res = 0;
        MezzmoContext * mezzmoCtx = ctx->priv;
    int16_t b_xl, b_yu, b_wd, b_ht;
    unsigned char bgc[4];
    int    valFade = 0;
    int    jobFade = 0;
    int    valRoll = 0;
    int    jobRoll = 0;
    int    posItem = 0;
    int v_wd = pic_ref->width;
    int v_ht = pic_ref->height;
    if (0 != mezzmoCtx->hsbs) v_wd = mezzmoCtx->hsbs;
    if (0 != mezzmoCtx->vsbs) v_ht = mezzmoCtx->vsbs;
        // 
    if (0 == mezzmoCtx->fifoDone) return 0;
    if ((v_wd > 32767) || (v_wd < 16)) return 0;
    if ((v_ht > 32767) || (v_ht < 16)) return 0;
    if (!load_fifo(mezzmoCtx, pic_ref)) return 0;
    for (posItem = 0; posItem < MAX_MEZZMOITEM; posItem++ )
    {
        if (mezzmoCtx->mezzmoItem[posItem].area)
        {
            if (vpts >= mezzmoCtx->mezzmoItem[posItem].ptsS)
            {
                if (vpts <= mezzmoCtx->mezzmoItem[posItem].ptsE)
                {
                    b_xl = mezzmoCtx->mezzmoItem[posItem].b_xl;
                    b_yu = mezzmoCtx->mezzmoItem[posItem].b_yu;
                    b_wd = mezzmoCtx->mezzmoItem[posItem].b_wd;
                    b_ht = mezzmoCtx->mezzmoItem[posItem].b_ht;
                    valFade = mezzmoCtx->mezzmoItem[posItem].fade;
                    jobFade = mezzmoCtx->mezzmoItem[posItem].jobF;
                    valRoll = mezzmoCtx->mezzmoItem[posItem].roll;
                    jobRoll = mezzmoCtx->mezzmoItem[posItem].jobR;
                    if (valRoll > 0)
                    {
                        if ((1 == valRoll) || (2 == valRoll)) b_xl = jobRoll;
                        if ((3 == valRoll) || (4 == valRoll)) b_yu = jobRoll;
                    }
                    if (0x00 != mezzmoCtx->bkgr_yuva[3])
                    {
                        bgc[0] = mezzmoCtx->bkgr_yuva[0];
                        bgc[1] = mezzmoCtx->bkgr_yuva[1];
                        bgc[2] = mezzmoCtx->bkgr_yuva[2];
                        bgc[3] = (mezzmoCtx->bkgr_yuva[3] > jobFade) ? mezzmoCtx->bkgr_yuva[3]-jobFade : 0;
                        if (0x00 != bgc[3])
                        {
                            draw_box( pic_ref, v_wd, v_ht, b_xl, b_yu, b_wd+8, b_ht, bgc, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, 0 );
                            if (0 != mezzmoCtx->hsbs) draw_box( pic_ref, v_wd, v_ht, b_xl, b_yu, b_wd+8, b_ht, bgc, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, mezzmoCtx->hsbs, 0 );
                            if (0 != mezzmoCtx->vsbs) draw_box( pic_ref, v_wd, v_ht, b_xl, b_yu, b_wd+8, b_ht, bgc, 1, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, mezzmoCtx->vsbs );
                        }
                    }
                    res = draw_img( pic_ref, v_wd, v_ht, b_xl, b_yu, b_wd, b_ht, mezzmoCtx->mezzmoItem[posItem].area, jobFade, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, 0 );        
                    if (0 != mezzmoCtx->hsbs) draw_img( pic_ref, v_wd, v_ht, b_xl, b_yu, b_wd, b_ht, mezzmoCtx->mezzmoItem[posItem].area, jobFade, mezzmoCtx->hsub, mezzmoCtx->vsub, mezzmoCtx->hsbs, 0 );
                    if (0 != mezzmoCtx->vsbs) draw_img( pic_ref, v_wd, v_ht, b_xl, b_yu, b_wd, b_ht, mezzmoCtx->mezzmoItem[posItem].area, jobFade, mezzmoCtx->hsub, mezzmoCtx->vsub, 0, mezzmoCtx->vsbs );
                    if (valFade > 0)
                    {            
                        if (((255 - valFade) * 10) > (vpts - mezzmoCtx->mezzmoItem[posItem].ptsS))
                        {
                            jobFade -= valFade; 
                            if (jobFade < 0) jobFade = 0;
                        }
                        else if (((255 - valFade) * 10) > (mezzmoCtx->mezzmoItem[posItem].ptsE - vpts))
                        {
                            jobFade += valFade;
                            if (jobFade > 255) jobFade = 255;
                        }
                        else jobFade = 0;
                        mezzmoCtx->mezzmoItem[posItem].jobF = jobFade;
                    }
                    if (valRoll > 0)
                    {
                             if (1 == valRoll) { jobRoll += 1; if (jobRoll > v_wd) valRoll = 0; }
                        else if (2 == valRoll) { jobRoll -= 1; if (jobRoll < 0)    valRoll = 0; }
                        else if (3 == valRoll) { jobRoll += 1; if (jobRoll > v_ht) valRoll = 0; }
                        else if (4 == valRoll) { jobRoll -= 1; if (jobRoll < 0)    valRoll = 0; }
                        mezzmoCtx->mezzmoItem[posItem].jobR = jobRoll;
                    }
                }
                else
                {
                    av_free( mezzmoCtx->mezzmoItem[posItem].area );
                    mezzmoCtx->mezzmoItem[posItem].area = 0;
                }
            }
        }
    }
    return res;
}



static int filter_frame(AVFilterLink *inlink, AVFrame *frame)
{
    AVFilterContext   * ctx     = inlink->dst;
    AVFilterLink      * outlink  = inlink->dst->outputs[0];
    MezzmoContext * mezzmoCtx = ctx->priv;
    int64_t vpts;
    uint8_t isDraw = 0;
    double timeBase = av_q2d(inlink->time_base);
    vpts = (frame->pts * 1000) * timeBase; 
    switch (mezzmoCtx->dSbs)
    {
    case 1:    mezzmoCtx->hsbs = frame->width >> 1;  break;
    case 2:    mezzmoCtx->vsbs = frame->height >> 1;  break;
    }
    isDraw |= process_fifo( ctx, frame, vpts, timeBase );
    isDraw |= process_ibox( ctx, frame, vpts, timeBase );
    return ff_filter_frame( outlink, frame );
}


AVFILTER_DEFINE_CLASS(mezzmo);


AVFilter ff_vf_mezzmo = {
    .name          = "mezzmo",
    .description   = NULL_IF_CONFIG_SMALL("Mezzmo filtering."),
    .priv_size     = sizeof(MezzmoContext),
    .priv_class    = &mezzmo_class,
    .init          = init,
    .uninit        = uninit,
    .query_formats = query_formats,
    .inputs        = (const AVFilterPad[]) {{ .name             = "default",
                                    .type             = AVMEDIA_TYPE_VIDEO,
                                    .get_video_buffer = ff_null_get_video_buffer,
                                    .filter_frame     = filter_frame,
                                    .config_props     = config_input,
                                    .min_perms        = AV_PERM_WRITE | AV_PERM_READ },
                                  { .name = NULL}},
    .outputs   = (const AVFilterPad[]) {{ .name             = "default",
                                    .type             = AVMEDIA_TYPE_VIDEO, },
                                  { .name = NULL}},
};

//--EndOfFile-----------------------------------------------------------------------------



